@@ -1,6 +1,6 @@
 import json
 import logging
-from typing import Callable, Dict, Hashable, List, Optional
+from typing import Callable, Hashable, Optional
 
 import numpy as np
 import rdkit
@@ -11,7 +11,7 @@ from tqdm import tqdm
 from nfp.preprocessing import features
 from nfp.preprocessing.features import Tokenizer
 
-zero = tf.constant(0, dtype=tf.int64)
+logger = logging.getLogger(__name__)
 
 
 class MolPreprocessor(object):
@@ -23,6 +23,8 @@ class MolPreprocessor(object):
         atom_features: A function applied to an rdkit.Atom that returns some
             representation (i.e., string, integer) for the Tokenizer class.
         bond_features: A function applied to an rdkit Bond to return some description.
+        output_dtype: The datatype for the arrays returned by `construct_feature_matrices`
+        bond_indices: Whether to return bond_incides in `construct_feature_matrices`
 
     :Example:
     >>> preprocessor = MolPreprocessor(explicit_hs=False)
@@ -37,7 +39,10 @@ class MolPreprocessor(object):
 
     def __init__(self,
                  atom_features: Optional[Callable[[rdkit.Chem.Atom], Hashable]] = None,
-                 bond_features: Optional[Callable[[rdkit.Chem.Bond], Hashable]] = None) -> None:
+                 bond_features: Optional[Callable[[rdkit.Chem.Bond], Hashable]] = None,
+                 output_dtype: str = 'int32',
+                 bond_indices: bool = False,
+                 ) -> None:
 
         self.atom_tokenizer = Tokenizer()
         self.bond_tokenizer = Tokenizer()
@@ -50,6 +55,8 @@ class MolPreprocessor(object):
 
         self.atom_features = atom_features
         self.bond_features = bond_features
+        self.output_dtype = output_dtype
+        self.bond_indices = bond_indices
 
         # Keep track of biggest molecules seen in training
         self.max_atoms = 0
@@ -60,26 +67,38 @@ class MolPreprocessor(object):
             return json.dump(self, f, default=lambda x: x.__dict__)
 
     def from_json(self, filename):
+
         with open(filename, 'r') as f:
             json_data = json.load(f)
 
         load_from_json(self, json_data)
 
     @property
-    def atom_classes(self):
+    def atom_classes(self) -> int:
         """ The number of atom types found (includes the 0 null-atom type) """
         return self.atom_tokenizer.num_classes + 1
 
     @property
-    def bond_classes(self):
+    def bond_classes(self) -> int:
         """ The number of bond types found (includes the 0 null-bond type) """
         return self.bond_tokenizer.num_classes + 1
 
-    def construct_feature_matrices(self, mol: rdkit.Chem.Mol, train: bool = False) -> {}:
+    def construct_feature_matrices(self,
+                                   mol: rdkit.Chem.Mol,
+                                   train: bool = False,
+                                   max_num_atoms: Optional[int] = None,
+                                   max_num_bonds: Optional[int] = None,
+                                   ) -> {str: np.ndarray}:
         """ Convert an rdkit Mol to a list of tensors
-        'atom' : (n_atom,) length list of atom classes
-        'bond' : (n_bond,) list of bond classes
-        'connectivity' : (n_bond, 2) array of source atom, target atom pairs.
+
+        Parameters
+        ----------
+        mol : rdkit.Chem.Mol
+        train : bool
+        max_num_atoms : int, optional
+            Specify the size of the output arrays with a maximum number of atoms
+        max_num_bonds : int, optional
+            Maximum number of bonds in the output array
         """
 
         self.atom_tokenizer.train = train
@@ -92,9 +111,13 @@ class MolPreprocessor(object):
         if n_bond == 0:
             n_bond = 1
 
-        atom_feature_matrix = np.zeros(n_atom, dtype='int32')
-        bond_feature_matrix = np.zeros(n_bond, dtype='int32')
-        connectivity = np.zeros((n_bond, 2), dtype='int32')
+        max_num_atoms = mol.GetNumAtoms() if max_num_atoms is None else max_num_atoms
+        max_num_bonds = n_bond if max_num_bonds is None else max_num_bonds
+
+        atom_feature_matrix = np.zeros(max_num_atoms, dtype=self.output_dtype)
+        bond_feature_matrix = np.zeros(max_num_bonds, dtype=self.output_dtype)
+        connectivity = np.zeros((max_num_bonds, 2), dtype=self.output_dtype)
+        bond_indices = np.zeros(max_num_bonds, dtype=self.output_dtype)
 
         if n_bond == 1:
             bond_feature_matrix[0] = self.bond_tokenizer('self-link')
@@ -114,6 +137,9 @@ class MolPreprocessor(object):
                 # Bond Classes
                 bond_feature_matrix[bond_index] = self.bond_tokenizer(self.bond_features(bond, flipped=rev))
 
+                # Connect edges to original bonds
+                bond_indices[bond_index] = bond.GetIdx()
+
                 # Connectivity
                 if not rev:  # Original direction
                     connectivity[bond_index, 0] = bond.GetBeginAtomIdx()
@@ -125,23 +151,39 @@ class MolPreprocessor(object):
 
                 bond_index += 1
 
-        return {
+        output = {
             'atom': atom_feature_matrix,
             'bond': bond_feature_matrix,
             'connectivity': connectivity,
         }
 
-    output_signature = {'atom': tf.TensorSpec(shape=(None,), dtype=tf.int32),
-                        'bond': tf.TensorSpec(shape=(None,), dtype=tf.int32),
-                        'connectivity': tf.TensorSpec(shape=(None, 2), dtype=tf.int32)}
+        if self.bond_indices:
+            output['bond_indices'] = bond_indices
+
+        return output
+
+    @property
+    def output_signature(self) -> {str: tf.TensorSpec}:
+        signature = {'atom': tf.TensorSpec(shape=(None,), dtype=self.output_dtype),
+                     'bond': tf.TensorSpec(shape=(None,), dtype=self.output_dtype),
+                     'connectivity': tf.TensorSpec(shape=(None, 2), dtype=self.output_dtype)}
+
+        if self.bond_indices:
+            signature['bond_indices'] = tf.TensorSpec(shape=(None,), dtype=self.output_dtype)
+
+        return signature
 
 
 def load_from_json(obj, data):
     for key, val in obj.__dict__.items():
-        if type(val) == type(data[key]):
-            obj.__dict__[key] = data[key]
-        elif hasattr(val, '__dict__'):
-            load_from_json(val, data[key])
+        try:
+            if isinstance(val, type(data[key])):
+                obj.__dict__[key] = data[key]
+            elif hasattr(val, '__dict__'):
+                load_from_json(val, data[key])
+
+        except KeyError:
+            logger.warning(f"{key} not found in JSON file, it may have been created with an older nfp version")
 
 
 class SmilesPreprocessor(MolPreprocessor):
@@ -150,12 +192,11 @@ class SmilesPreprocessor(MolPreprocessor):
         super(SmilesPreprocessor, self).__init__(*args, **kwargs)
         self.explicit_hs = explicit_hs
 
-    def construct_feature_matrices(self, smiles: str, train: bool = False) -> {}:
+    def construct_feature_matrices(self, smiles: str, train: bool = False, **kwargs) -> {}:
         mol = rdkit.Chem.MolFromSmiles(smiles)
         if self.explicit_hs:
             mol = rdkit.Chem.AddHs(mol)
-        return super(SmilesPreprocessor, self).construct_feature_matrices(mol, train=train)
-
+        return super(SmilesPreprocessor, self).construct_feature_matrices(mol, train=train, **kwargs)
 
 
 def get_max_atom_bond_size(smiles_iterator, explicit_hs=True):
@@ -180,19 +221,3 @@ def canonicalize_smiles(smiles, isomeric=True, sanitize=True):
         return MolToSmiles(mol, isomericSmiles=isomeric)
     except Exception:
         pass
-
-
-def filter_keys(input_dict: Dict, keys: Optional[List[str]] = None) -> Dict:
-    """Remove unnecessary model inputs from nfp.SmilesPreprocessor outputs
-
-    Args:
-        keys: desired keys
-        input_dict: A dictionary containing unnecessary keys
-
-    Returns:
-        attribute_dict: The input dictionary filtered to the desired keys
-    """
-    if keys is None:
-        keys = ['atom', 'bond', 'connectivity']
-
-    return {key: value for key, value in input_dict.items() if key in keys}
